@@ -14,20 +14,24 @@
 #include "phosh-osk-enums.h"
 #include "pos-emoji-picker.h"
 #include "pos-input-method.h"
+#include "pos-clipboard-manager.h"
 #include "pos-completer.h"
 #include "pos-completer-manager.h"
 #include "pos-completion-bar.h"
 #include "pos-input-surface.h"
+#include "pos-logind-session.h"
 #include "pos-main.h"
 #include "pos-osk-widget.h"
+#include "pos-settings-panel.h"
 #include "pos-shortcuts-bar.h"
 #include "pos-vk-driver.h"
 #include "pos-virtual-keyboard.h"
 #include "pos-vk-driver.h"
 #include "util.h"
 
+#include <gmobile.h>
+
 #include <handy.h>
-#define LIBFEEDBACK_USE_UNSTABLE_API
 #include <libfeedback.h>
 
 #define GNOME_DESKTOP_USE_UNSTABLE_API
@@ -58,6 +62,7 @@ enum {
   PROP_INPUT_METHOD,
   PROP_COMPLETER,
   PROP_COMPLETER_MANAGER,
+  PROP_CLIPBOARD_MANAGER,
   PROP_SCREEN_KEYBOARD_ENABLED,
   PROP_KEYBOARD_DRIVER,
   PROP_SURFACE_VISIBLE,
@@ -98,7 +103,9 @@ struct _PosInputSurface {
   GSettings               *osk_settings;
   GnomeXkbInfo            *xkbinfo;
 
-  /* wayland input-method */
+  PosLogindSession        *logind_session;
+
+  /* Wayland input-method */
   PosInputMethod          *input_method;
 
   /* OSK */
@@ -129,6 +136,12 @@ struct _PosInputSurface {
   gboolean                 completion_enabled;
   PhoshOskCompletionModeFlags completion_mode;
 
+  /* Clipboard */
+  PosClipboardManager    *clipboard_manager;
+
+  /* Swipe gesture */
+  GtkGesture              *swipe_down;
+
   /* emission hook for clicks */
   gulong                   clicked_id;
 };
@@ -142,6 +155,24 @@ G_DEFINE_TYPE_WITH_CODE (PosInputSurface, pos_input_surface, PHOSH_TYPE_LAYER_SU
                          G_IMPLEMENT_INTERFACE (G_TYPE_ACTION_MAP, pos_input_surface_action_map_iface_init)
   )
 
+#define MIN_Y_VELOCITY 1000
+
+static void
+on_swipe (GtkGestureSwipe *swipe, double velocity_x, double velocity_y, gpointer data)
+{
+  PosInputSurface *self = POS_INPUT_SURFACE (data);
+
+  g_return_if_fail (GTK_IS_GESTURE_SWIPE (swipe));
+
+  g_debug ("swipe with v_x: %f, v_y: %f", velocity_x, velocity_y);
+
+  if (velocity_y > MIN_Y_VELOCITY && velocity_y > 2.0 * ABS (velocity_x)) {
+    g_debug ("Hiding the keyboard on swipe down");
+    pos_input_surface_set_visible (self, FALSE);
+  } else {
+    g_debug ("Swipe not downwards");
+  }
+}
 
 static void
 on_shortcut_activated (PosInputSurface *self, PosShortcut *shortcut, PosShortcutsBar *bar)
@@ -226,7 +257,7 @@ pos_input_surface_is_completion_mode (PosInputSurface *self)
 static void
 on_completion_selected (PosInputSurface *self, const char *completion)
 {
-  g_autofree gchar *send = NULL;
+  g_autofree char *send = NULL;
 
   g_return_if_fail (POS_IS_INPUT_SURFACE (self));
   g_return_if_fail (completion != NULL);
@@ -236,8 +267,11 @@ on_completion_selected (PosInputSurface *self, const char *completion)
 
   pos_input_method_send_string (self->input_method, send, TRUE);
 
-  if (pos_input_surface_is_completer_active (self))
+  if (pos_input_surface_is_completer_active (self)) {
+    pos_completer_learn_accepted (self->completer, send);
     pos_completer_set_preedit (self->completer, NULL);
+  }
+
 }
 
 
@@ -363,7 +397,7 @@ on_osk_key_symbol (PosInputSurface *self, const char *symbol, GtkWidget *osk_wid
 
   /* virtual-keyboard, no input method */
   if (!pos_input_method_get_active (self->input_method)) {
-    pos_vk_driver_key_down (self->keyboard_driver, symbol);
+    pos_vk_driver_key_down (self->keyboard_driver, symbol, POS_KEYCODE_MODIFIER_NONE);
     pos_vk_driver_key_up (self->keyboard_driver, symbol);
     return;
   }
@@ -375,7 +409,7 @@ on_osk_key_symbol (PosInputSurface *self, const char *symbol, GtkWidget *osk_wid
   }
 
   if (g_str_has_prefix (symbol, "KEY_")) {
-    pos_vk_driver_key_down (self->keyboard_driver, symbol);
+    pos_vk_driver_key_down (self->keyboard_driver, symbol, POS_KEYCODE_MODIFIER_NONE);
     pos_vk_driver_key_up (self->keyboard_driver, symbol);
   } else {
     pos_input_method_send_string (self->input_method, symbol, TRUE);
@@ -464,8 +498,8 @@ clipboard_copy_activated (GSimpleAction *action,
 {
   PosInputSurface *self = POS_INPUT_SURFACE (data);
 
-  pos_vk_driver_key_down (self->keyboard_driver, "KEY_COPY");
-  pos_vk_driver_key_up (self->keyboard_driver, "KEY_COPY");
+  pos_vk_driver_key_down (self->keyboard_driver, "c", POS_KEYCODE_MODIFIER_CTRL);
+  pos_vk_driver_key_up (self->keyboard_driver, "c");
 }
 
 
@@ -475,13 +509,37 @@ clipboard_paste_activated (GSimpleAction *action,
                            gpointer       data)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (data);
+  const char *text;
 
-  pos_vk_driver_key_down (self->keyboard_driver, "KEY_PASTE");
-  pos_vk_driver_key_up (self->keyboard_driver, "KEY_PASTE");
+  text = pos_clipboard_manager_get_text (self->clipboard_manager);
+  if (gm_str_is_null_or_empty (text))
+    return;
+
+  if (pos_input_method_get_active (self->input_method)) {
+    pos_input_surface_submit_current_preedit (self);
+    pos_input_method_send_string (self->input_method, text, TRUE);
+  } else {
+    /* TODO */
+    g_warning_once ("Pasting via vk-driver not yet supported");
+  }
+
+  /* Close popover in case we pasted from there */
+  gtk_popover_popdown (self->menu_popup);
 }
 
-/* Emoji picker */
 
+static void
+settings_activated (GSimpleAction *action, GVariant *parameter, gpointer data)
+{
+  PosInputSurface *self = POS_INPUT_SURFACE (data);
+
+  /* popdown popover right away to avoid flicker when OSK goes away */
+  gtk_popover_popdown (self->menu_popup);
+  pos_open_settings_panel ("osk");
+}
+
+
+/* Emoji picker */
 
 static void
 send_emoji_via_vk (PosInputSurface *self, const char *emoji)
@@ -507,7 +565,7 @@ send_emoji_via_vk (PosInputSurface *self, const char *emoji)
   for (int i = 0; syms_array->pdata[i]; i++) {
     const char *symbol = syms_array->pdata[i];
 
-    pos_vk_driver_key_down (self->keyboard_driver, symbol);
+    pos_vk_driver_key_down (self->keyboard_driver, symbol, POS_KEYCODE_MODIFIER_NONE);
     pos_vk_driver_key_up (self->keyboard_driver, symbol);
   }
 
@@ -652,8 +710,10 @@ select_layout_change_state (GSimpleAction *action,
   if (osk_widget == NULL) {
     if (g_str_equal (layout, "terminal")) {
       osk_widget = self->osk_terminal;
+      pos_input_surface_submit_current_preedit (self);
     } else if (g_str_equal (layout, "emoji")) {
       osk_widget = self->emoji_picker;
+      pos_input_surface_submit_current_preedit (self);
     } else {
       g_warning ("Failed to find layout '%s'", layout);
       return;
@@ -868,6 +928,26 @@ pos_input_surface_set_completer_manager (PosInputSurface     *self,
 }
 
 
+static void
+pos_input_surface_set_clipboard_manager (PosInputSurface     *self,
+                                         PosClipboardManager *clipboard_manager)
+{
+  GAction *paste_action;
+
+  if (self->clipboard_manager == clipboard_manager)
+    return;
+
+  g_set_object (&self->clipboard_manager, clipboard_manager);
+
+  paste_action = g_action_map_lookup_action (G_ACTION_MAP (self), "clipboard-paste");
+  g_assert (paste_action);
+
+  g_object_bind_property (clipboard_manager, "has-text",
+                          paste_action, "enabled",
+                          G_BINDING_SYNC_CREATE);
+}
+
+
 static double
 reverse_ease_out_cubic (double t)
 {
@@ -929,6 +1009,9 @@ pos_input_surface_set_property (GObject      *object,
   case PROP_COMPLETER_MANAGER:
     pos_input_surface_set_completer_manager (self, g_value_get_object (value));
     break;
+  case PROP_CLIPBOARD_MANAGER:
+    pos_input_surface_set_clipboard_manager (self, g_value_get_object (value));
+    break;
   case PROP_SCREEN_KEYBOARD_ENABLED:
     pos_screen_keyboard_set_enabled (self, g_value_get_boolean (value));
     break;
@@ -966,6 +1049,9 @@ pos_input_surface_get_property (GObject    *object,
   case PROP_COMPLETER_MANAGER:
     g_value_set_object (value, self->completer_manager);
     break;
+  case PROP_CLIPBOARD_MANAGER:
+    g_value_set_object (value, self->clipboard_manager);
+    break;
   case PROP_SCREEN_KEYBOARD_ENABLED:
     g_value_set_boolean (value, pos_input_surface_get_screen_keyboard_enabled (self));
     break;
@@ -1086,7 +1172,8 @@ on_im_surrounding_text_changed (PosInputSurface *self, GParamSpec *pspec, PosInp
   g_assert (POS_IS_INPUT_METHOD (im));
 
   text = pos_input_method_get_surrounding_text (im, &anchor, &cursor);
-  if (!pos_input_surface_is_completer_active (self))
+
+  if (!pos_input_surface_is_completion_mode (self))
     return;
 
   before = g_strndup (text, cursor);
@@ -1134,7 +1221,6 @@ static void
 pos_input_surface_constructed (GObject *object)
 {
   static gboolean input_serial_sent = FALSE;
-
   PosInputSurface *self = POS_INPUT_SURFACE (object);
   const char *test_layout = g_getenv ("POS_TEST_LAYOUT");
 
@@ -1198,6 +1284,7 @@ pos_input_surface_finalize (GObject *object)
 
   g_clear_handle_id (&self->animation.id, g_source_remove);
 
+  g_clear_object (&self->logind_session);
   g_clear_object (&self->keyboard_driver);
   g_clear_object (&self->input_method);
   g_clear_object (&self->a11y_settings);
@@ -1205,8 +1292,10 @@ pos_input_surface_finalize (GObject *object)
   g_clear_object (&self->osk_settings);
   g_clear_object (&self->xkbinfo);
   g_clear_object (&self->css_provider);
+  g_clear_object (&self->clipboard_manager);
   g_clear_object (&self->completer);
   g_clear_object (&self->completer_manager);
+  g_clear_object (&self->swipe_down);
   g_clear_pointer (&self->theme_name, g_free);
   g_clear_pointer (&self->osks, g_hash_table_destroy);
 
@@ -1214,7 +1303,7 @@ pos_input_surface_finalize (GObject *object)
 }
 
 
-static gchar **
+static char **
 pos_input_surface_list_actions (GActionGroup *group)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (group);
@@ -1228,7 +1317,7 @@ pos_input_surface_list_actions (GActionGroup *group)
 
 static gboolean
 pos_input_surface_query_action (GActionGroup        *group,
-                                const gchar         *action_name,
+                                const char          *action_name,
                                 gboolean            *enabled,
                                 const GVariantType **parameter_type,
                                 const GVariantType **state_type,
@@ -1247,7 +1336,7 @@ pos_input_surface_query_action (GActionGroup        *group,
 
 static void
 pos_input_surface_activate_action (GActionGroup *group,
-                                   const gchar  *action_name,
+                                   const char   *action_name,
                                    GVariant     *parameter)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (group);
@@ -1261,7 +1350,7 @@ pos_input_surface_activate_action (GActionGroup *group,
 
 static void
 pos_input_surface_change_action_state (GActionGroup *group,
-                                       const gchar  *action_name,
+                                       const char   *action_name,
                                        GVariant     *state)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (group);
@@ -1284,7 +1373,7 @@ pos_input_surface_action_group_iface_init (GActionGroupInterface *iface)
 
 
 static GAction *
-pos_input_surface_lookup_action (GActionMap  *action_map, const gchar *action_name)
+pos_input_surface_lookup_action (GActionMap  *action_map, const char *action_name)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (action_map);
 
@@ -1306,7 +1395,7 @@ pos_input_surface_add_action (GActionMap *action_map, GAction *action)
 }
 
 static void
-pos_input_surface_remove_action (GActionMap *action_map, const gchar *action_name)
+pos_input_surface_remove_action (GActionMap *action_map, const char *action_name)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (action_map);
 
@@ -1379,9 +1468,9 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, completion_bar);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, deck);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, emoji_picker);
-  gtk_widget_class_bind_template_child (widget_class, PosInputSurface, osk_terminal);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, menu_box_layouts);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, menu_popup);
+  gtk_widget_class_bind_template_child (widget_class, PosInputSurface, osk_terminal);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, shortcuts_bar);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, word_completion_btn);
   gtk_widget_class_bind_template_callback (widget_class, on_completion_selected);
@@ -1427,7 +1516,17 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
     g_param_spec_object ("completer-manager", "", "",
                          POS_TYPE_COMPLETER_MANAGER,
                          G_PARAM_READWRITE |
-                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+  /**
+   * PosInputSurface:clipboard-manager:
+   *
+   * A clipboard manager to use. The clipboard manager is used to get text to paste.
+   */
+  props[PROP_CLIPBOARD_MANAGER] =
+    g_param_spec_object ("clipboard-manager", "", "",
+                         POS_TYPE_CLIPBOARD_MANAGER,
+                         G_PARAM_READWRITE |
+                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
   /**
    * PosInputSurface:enable
    *
@@ -1564,9 +1663,9 @@ static PosOskWidget *
 insert_xkb_layout (PosInputSurface *self, const char *type, const char *layout_id)
 {
   g_autofree char *name = NULL;
-  const gchar *layout = NULL;
-  const gchar *variant = NULL;
-  const gchar *display_name = NULL;
+  const char *layout = NULL;
+  const char *variant = NULL;
+  const char *display_name = NULL;
 
   if (g_strcmp0 (type, "xkb")) {
     g_debug ("Not a xkb layout: '%s' - ignoring", layout_id);
@@ -1722,6 +1821,7 @@ static GActionEntry entries[] =
 {
   { .name = "clipboard-copy", .activate = clipboard_copy_activated },
   { .name = "clipboard-paste", .activate = clipboard_paste_activated },
+  { .name = "settings", .activate = settings_activated },
   { .name = "select-layout", .parameter_type = "s", .state = "\"terminal\"",
     .change_state = select_layout_change_state },
   { .name = "menu", .parameter_type = "(ii)", .activate = menu_activated },
@@ -1734,6 +1834,7 @@ pos_input_surface_init (PosInputSurface *self)
   GtkSettings *gtk_settings;
   g_autoptr (GPropertyAction) completion_action = NULL;
   g_autoptr (GError) err = NULL;
+  GAction *action;
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
@@ -1790,6 +1891,19 @@ pos_input_surface_init (PosInputSurface *self)
 
   /* Disable swipe gestures to change layouts by default */
   pos_input_surface_set_layout_swipe (self, FALSE);
+
+  self->swipe_down = g_object_new (GTK_TYPE_GESTURE_SWIPE,
+                                   "widget", self,
+                                   "propagation-phase", GTK_PHASE_CAPTURE,
+                                   "touch-only", TRUE,
+                                   NULL);
+  g_signal_connect (self->swipe_down, "swipe", G_CALLBACK (on_swipe), self);
+
+  self->logind_session = pos_logind_session_new ();
+  action = g_action_map_lookup_action (G_ACTION_MAP (self->action_map), "settings");
+  g_object_bind_property (self->logind_session, "locked",
+                          action, "enabled",
+                          G_BINDING_SYNC_CREATE | G_BINDING_INVERT_BOOLEAN);
 }
 
 
